(function() {
  var Player, fs, retro,
    bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  retro = require('node-retro');

  fs = require('fs');

  AudioBuffer.prototype.copyToChannel = function(source, channelNumber, startInChannel) {
    return this.getChannelData(channelNumber | 0).set(source, startInChannel | 0);
  };

  module.exports = Player = (function() {
    Player.prototype.pixelFormat = retro.PIXEL_FORMAT_0RGB1555;

    Player.prototype.variables = {};

    Player.prototype.romTemp = 'temp.rom';

    Player.prototype.variablesUpdate = false;

    Player.prototype.overscan = false;

    Player.prototype.running = false;

    function Player(gl, audio, input, core1, game, save) {
      var i;
      this.gl = gl;
      this.audio = audio;
      this.input = input;
      this.core = core1;
      this.game = game;
      this.save = save;
      this.frame = bind(this.frame, this);
      this.environment = bind(this.environment, this);
      this.audiosamplebatch = bind(this.audiosamplebatch, this);
      this.videorefresh = bind(this.videorefresh, this);
      this.initGL();
      this.info = this.core.getSystemInfo();
      if (typeof this.game === 'string') {
        this.core.loadGamePath(this.game);
      } else {
        if (this.info.need_fullpath) {
          fs.writeFileSync(this.romTemp, this.game);
          this.core.loadGamePath(this.romTemp);
        } else {
          this.core.loadGame(this.game);
        }
      }
      if (this.save != null) {
        this.core.unserialize(this.save);
      }
      this.av_info = this.core.getSystemAVInfo();
      this.fpsInterval = 1000 / this.av_info.timing.fps;
      this.then = 0;
      this.sampleRate = this.av_info.timing.sample_rate;
      this.bufferSize = 256;
      this.latency = 96;
      this.numBuffers = Math.floor(this.latency * this.sampleRate / (1000 * this.bufferSize));
      if (this.numBuffers < 2) {
        this.numBuffers = 2;
      }
      i = 0;
      this.buffers = [];
      while (i < this.numBuffers) {
        this.buffers[i] = this.audio.createBuffer(2, this.bufferSize, this.sampleRate);
        i++;
      }
      this.bufOffset = 0;
      this.bufIndex = 0;
      this.core.on('videorefresh', this.videorefresh);
      this.core.on('inputstate', this.input);
      this.core.on('audiosamplebatch', this.audiosamplebatch);
      this.core.on('log', this.log);
      this.core.on('environment', this.environment);
    }

    Player.prototype.initGL = function() {
      var buffer, fragmentShader, pArray, positionLocation, program, texArray, texCoordBuffer, texCoordLocation, vertexShader;
      fragmentShader = this.gl.createShader(this.gl.FRAGMENT_SHADER);
      this.gl.shaderSource(fragmentShader, 'precision mediump float; uniform sampler2D u_image; varying vec2 v_texCoord; void main() { gl_FragColor = texture2D(u_image, v_texCoord); }');
      this.gl.compileShader(fragmentShader);
      vertexShader = this.gl.createShader(this.gl.VERTEX_SHADER);
      this.gl.shaderSource(vertexShader, 'attribute vec2 a_texCoord; attribute vec2 a_position; varying vec2 v_texCoord; void main() { gl_Position = vec4(a_position, 0, 1); v_texCoord = a_texCoord; }');
      this.gl.compileShader(vertexShader);
      program = this.gl.createProgram();
      this.gl.attachShader(program, vertexShader);
      this.gl.attachShader(program, fragmentShader);
      this.gl.linkProgram(program);
      this.gl.useProgram(program);
      positionLocation = this.gl.getAttribLocation(program, 'a_position');
      buffer = this.gl.createBuffer();
      this.gl.bindBuffer(this.gl.ARRAY_BUFFER, buffer);
      pArray = new Float32Array([-1, -1, 1, -1, -1, 1, -1, 1, 1, -1, 1, 1]);
      this.gl.bufferData(this.gl.ARRAY_BUFFER, pArray, this.gl.STATIC_DRAW);
      this.gl.enableVertexAttribArray(positionLocation);
      this.gl.vertexAttribPointer(positionLocation, 2, this.gl.FLOAT, false, 0, 0);
      texCoordLocation = this.gl.getAttribLocation(program, 'a_texCoord');
      texCoordBuffer = this.gl.createBuffer();
      this.gl.bindBuffer(this.gl.ARRAY_BUFFER, texCoordBuffer);
      texArray = new Float32Array([0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 1]);
      this.gl.bufferData(this.gl.ARRAY_BUFFER, texArray, this.gl.STATIC_DRAW);
      this.gl.enableVertexAttribArray(texCoordLocation);
      this.gl.vertexAttribPointer(texCoordLocation, 2, this.gl.FLOAT, false, 0, 0);
      this.texture = this.gl.createTexture();
      this.gl.bindTexture(this.gl.TEXTURE_2D, this.texture);
      this.gl.texParameteri(this.gl.TEXTURE_2D, this.gl.TEXTURE_WRAP_S, this.gl.CLAMP_TO_EDGE);
      this.gl.texParameteri(this.gl.TEXTURE_2D, this.gl.TEXTURE_WRAP_T, this.gl.CLAMP_TO_EDGE);
      this.gl.texParameteri(this.gl.TEXTURE_2D, this.gl.TEXTURE_MIN_FILTER, this.gl.LINEAR);
      return this.gl.pixelStorei(this.gl.UNPACK_FLIP_Y_WEBGL, true);
    };

    Player.prototype.videorefresh = function(data, w, h) {
      var format, type;
      if (!this.width && !this.height) {
        this.width = w;
        this.height = h;
        this.gl.canvas.width = this.width;
        this.gl.canvas.height = this.height;
        this.gl.viewport(0, 0, this.width, this.height);
      }
      switch (this.pixelFormat) {
        case retro.PIXEL_FORMAT_0RGB1555:
          data = new Uint16Array(data.slice(0));
          format = this.gl.RGB;
          type = this.gl.UNSIGNED_SHORT_5_5_5_1;
          break;
        case retro.PIXEL_FORMAT_XRGB8888:
          data = new Uint8Array(data.slice(0));
          format = this.gl.RGBA;
          type = this.gl.UNSIGNED_BYTE;
          break;
        case retro.PIXEL_FORMAT_RGB565:
          data = new Uint16Array(data.slice(0));
          format = this.gl.RGB;
          type = this.gl.UNSIGNED_SHORT_5_6_5;
      }
      this.gl.texImage2D(this.gl.TEXTURE_2D, 0, format, w, h, 0, format, type, data);
      return this.gl.drawArrays(this.gl.TRIANGLES, 0, 6);
    };

    Player.prototype.audiosamplebatch = function(left, right, frames) {
      var buf, count, fill, i, source, startTime;
      i = 0;
      while (i < this.bufIndex) {
        if (this.buffers[i].endTime < this.audio.currentTime) {
          buf = this.buffers.splice(i, 1)[0];
          this.buffers[this.numBuffers - 1] = buf;
          i--;
          this.bufIndex--;
        }
        i++;
      }
      count = 0;
      while (frames) {
        fill = this.buffers[this.bufIndex].length - this.bufOffset;
        if (fill > frames) {
          fill = frames;
        }
        this.buffers[this.bufIndex].copyToChannel(new Float32Array(left, count * 4, fill), 0, this.bufOffset);
        this.buffers[this.bufIndex].copyToChannel(new Float32Array(right, count * 4, fill), 1, this.bufOffset);
        this.bufOffset += fill;
        count += fill;
        frames -= fill;
        if (this.bufOffset === this.bufferSize) {
          if (this.bufIndex === this.numBuffers - 1) {
            break;
          }
          if (this.bufIndex) {
            startTime = this.buffers[this.bufIndex - 1].endTime;
          } else {
            startTime = this.audio.currentTime;
          }
          this.buffers[this.bufIndex].endTime = startTime + this.buffers[this.bufIndex].duration;
          source = this.audio.createBufferSource();
          source.buffer = this.buffers[this.bufIndex];
          source.connect(this.audio.destination);
          source.start(startTime);
          this.bufIndex++;
          this.bufOffset = 0;
        }
      }
      return count;
    };

    Player.prototype.setVariable = function(key, value) {
      this.variables[core][key] = value;
      return this.variablesUpdate = true;
    };

    Player.prototype.log = function(level, msg) {
      return console.log(msg);
    };

    Player.prototype.environment = function(cmd, value) {
      switch (cmd) {
        case retro.ENVIRONMENT_GET_OVERSCAN:
          return this.overscan;
        case retro.ENVIRONMENT_GET_VARIABLE_UPDATE:
          if (this.variablesUpdate) {
            this.variablesUpdate = false;
            return true;
          }
          return false;
        case retro.ENVIRONMENT_SET_PIXEL_FORMAT:
          this.pixelFormat = value;
          return true;
        case retro.ENVIRONMENT_GET_CAN_DUPE:
          return true;
        case retro.ENVIRONMENT_GET_SYSTEM_DIRECTORY:
          return '.';
        case retro.ENVIRONMENT_GET_VARIABLE:
          return this.variables[value];
        case retro.ENVIRONMENT_SET_INPUT_DESCRIPTORS:
          return true;
        default:
          console.log("Unknown environment command " + cmd);
          return false;
      }
    };

    Player.prototype.frame = function(now) {
      var elapsed;
      if (!this.running) {
        return;
      }
      elapsed = now - this.then;
      if (elapsed > this.fpsInterval) {
        this.then = now - elapsed % this.fpsInterval;
        this.core.run();
      }
      return requestAnimationFrame(this.frame);
    };

    Player.prototype.start = function() {
      this.running = true;
      return this.frame();
    };

    Player.prototype.stop = function() {
      return this.running = false;
    };

    Player.prototype.deinit = function() {
      return this.stop();
    };

    return Player;

  })();

}).call(this);
